@@ -251,15 +251,23 @@ class AbstractExperiment:
     """
     Abstract base class for the experiment context manager.
     """
+
+    DEFAULT_TEMPLATES = {
+        'analysis.py': TEMPLATE_ENV.get_template('analysis.py.j2'),
+        'annotations.rst': TEMPLATE_ENV.get_template('annotations.py.j2')
+    }
+
     def __init__(self,
                  base_path: str,
                  namespace: str,
                  glob: dict,
                  debug: bool = False,
+                 templates: t.Dict[str, j2.Template] = DEFAULT_TEMPLATES.copy()
                  ):
         self.base_path = base_path
         self.namespace = namespace
         self.glob = glob
+        self.templates = templates
 
         # Within an experiment it is strongly encouraged to do all printing to the console via the internal
         # "e.info" method, which is actually calling this logger instance. Using a logger has the
@@ -281,13 +289,6 @@ class AbstractExperiment:
 
         self.hooks: t.Dict[str, t.List[t.Callable]] = defaultdict(list)
 
-        # After the main experiment context, there can be another context "e.analysis" and it is encouraged
-        # that this context is used to define the creation of important artifacts for the analysis and
-        # visualization of results.
-        # The RecordCode context manager will be able to copy all the code within that context and that code
-        # is then copied to the artifacts file within the archive folder.
-        self.analysis = RecordCode(initial_stack_index=2)
-
         # ExperimentExchange is actually a singleton, which means that each invocation of the constructor
         # returns the very same object instance. This object instance is basically a global state which is
         # accessable from anywhere. It is used to exchange(!) parameters for experiment inheritance, which
@@ -314,6 +315,27 @@ class AbstractExperiment:
         # the reproducibility of the snapshots.
         self.dependency_paths: t.Dict[str, str] = {}
 
+        # 13.02.2023
+        # Moved the base "analysis" functionality to this base class, because I realized that this
+        # functionality is more generic and should be available to SubExperiment as well for example.
+        # In essence, this is a context manager which uses magic to record the code string of all the code
+        # that is defined within.
+        # After the main experiment context, there can be another context "e.analysis" and it is encouraged
+        # that this context is used to define the creation of important artifacts for the analysis and
+        # visualization of results.
+        # The RecordCode context manager will be able to copy all the code within that context and that code
+        # is then copied to the artifacts file within the archive folder.
+        self.analysis = RecordCode(initial_stack_index=3)
+
+        # What's happening here? Basically at the point where the analysis context manager gets
+        # executed the experiment itself is already finalized (__exit__ was already called) but
+        # after the analysis is finished we still somehow need to create the "analysis.py" file
+        # with the contents of said analysis context.
+        # This line accomplishes that. Upon calling its own __exit__, the RecordCode object will
+        # call all the callback functions added to "exit_callback" so we simpl put the method
+        # which renders the "analysis.py" file from the jinja template as one such callback!
+        self.analysis.exit_callbacks.append(lambda rc, i: self.render_template('analysis.py'))
+
     @property
     def p(self):
         """
@@ -405,6 +427,7 @@ class AbstractExperiment:
             'log_path': self.log_path,
             'code_path': self.code_path,
             'error': self.error,
+            'logger': self.logger,
         }
 
     def discover_parameters(self) -> None:
@@ -521,6 +544,14 @@ class AbstractExperiment:
             self.data['artifacts']['source'] = str(source_path)
             shutil.copy(source_path, self.code_path)
 
+    def render_template(self, file_name: str, context: dict = {}):
+        template = self.templates[file_name]
+        path = os.path.join(self.path, file_name)
+        self.data['artifacts']['templates'][file_name] = path
+        with open(path, mode='w') as file:
+            content = template.render(experiment=self, **context)
+            file.write(content)
+
     # -- Logging --
 
     def prepare_logger(self) -> None:
@@ -545,6 +576,12 @@ class AbstractExperiment:
             handler.setFormatter(formatter)
             self.logger.addHandler(handler)
 
+        # 13.02.2023
+        # Since the base "analysis" functionality now has been moved into this class and since that also
+        # supports a logger, we also pass on the reference of the newly initialized logger so that all the
+        # log messaged generated by the RecordCode object are also logged to the main stream.
+        self.analysis.logger = self.logger
+
     def info(self, message: str) -> None:
         """
         Logs the given ``message`` string as an "INFO" level log message.
@@ -649,16 +686,40 @@ class AbstractExperiment:
 
         return decorator
 
+    def set_parameter(self, parameter_name: str, parameter_value: t.Any) -> None:
+        """
+        This function sets the parameter with the string ``parameter_name`` to ``parameter_value``
+
+        Whenever a parameter should be modified at runtime, it is strongly recommended that this method is
+        used because it will apply the changes to the internal parameters dict as well as the globals
+        dictionary!
+
+        :param parameter_name: The string name of the parameter variable to be modified
+        :param parameter_value: The new value to be assigned to that experiment parameter
+
+        :returns: None
+        """
+        self.parameters[parameter_name] = parameter_value
+        self.glob[parameter_name] = parameter_value
+
+    def apply_parameter_hooks(self) -> None:
+        """
+        This method applies *all* the parameter hooks for the experiment.
+
+        What is a parameter hook?
+
+        Besides the custom user-added hooks, there are default hooks which can be defined which are
+        filter hooks for all the variables which the experiment recognizes as "parameters" (upper case
+        global variables). Inside these default filter hooks it is for example possible to *modify*
+        instead of overwrite the default value provided by the parent experiment!
+        """
+        for parameter_name, parameter_value in self.parameters.items():
+            parameter_value = self.apply_hook(parameter_name, value=parameter_value, default=parameter_value)
+            self.set_parameter(parameter_name, parameter_value)
+
     # -- Magic Methods --
 
-    def __enter__(self):
-        # As the very first thing we need to check if the invocation of this experiment is as a parent
-        # experiment in an inheritance chain. This would be the case if the file path of the experiment
-        # module exists as a key within the global ExperimentExchange object. Because in that case it
-        # was put there by a child experiment with the intent of passing modified variables to this
-        # experiment.
-        # If that is the case, the globals dictionary will be updated with the new values that were passed
-        # in by the child experiment.
+    def check_experiment_exchange(self):
         module_path = self.glob['__file__']
         if module_path in self.experiment_exchange:
             other: dict = self.experiment_exchange[module_path]
@@ -672,6 +733,16 @@ class AbstractExperiment:
             # a multi threading scenario another parent module with the same path could now be invoked.
             self.experiment_exchange.release(module_path)
 
+    def __enter__(self):
+        # As the very first thing we need to check if the invocation of this experiment is as a parent
+        # experiment in an inheritance chain. This would be the case if the file path of the experiment
+        # module exists as a key within the global ExperimentExchange object. Because in that case it
+        # was put there by a child experiment with the intent of passing modified variables to this
+        # experiment.
+        # If that is the case, the globals dictionary will be updated with the new values that were passed
+        # in by the child experiment.
+        self.check_experiment_exchange()
+
         # This method actually initializes/calculates the values of all the specific experiment paths
         # based on the values "base_path" and "namespace".
         # You might be asking: "Why do it as a separate method and not directly in the constructor?"
@@ -681,6 +752,12 @@ class AbstractExperiment:
         self.discover_parameters()
         self.initialize_paths()
 
+        # 16.02.2023
+        # This method will apply all the hooks for the parameters. Essentially it is possible to register
+        # hooks with the exact names of parameters and these hooks will then be executed here at the very
+        # beginning of the experiment.
+        self.apply_parameter_hooks()
+
         # We are going to set this magic variable in the original experiment module. This can then be
         # used to identify whether a module is actually an experiment or not
         self.glob['__experiment__'] = self
@@ -830,9 +907,6 @@ class Experiment(AbstractExperiment):
         self.logger: Optional[logging.Logger] = None
         self.work_tracker = self.work_tracker_class(0)
 
-        self.analysis = RecordCode(initial_stack_index=2)
-        self.analysis.exit_callbacks.append(lambda rc, i: self.render_template('analysis.py'))
-
         # ~ Parsing command line arguments
         # TODO: I could introduce abstract base class and dependency inject this
         self.arg_parser = ExperimentArgParser(
@@ -1036,14 +1110,6 @@ class Experiment(AbstractExperiment):
         self.info(f"saved experiment error to: {self.error_path}")
         self.info("\n".join(tb))
 
-    def render_template(self, file_name):
-        template = self.templates[file_name]
-        path = os.path.join(self.path, file_name)
-        self.data['artifacts']['templates'][file_name] = path
-        with open(path, mode='w') as file:
-            content = template.render(experiment=self)
-            file.write(content)
-
     def render_templates(self):
         self.data['artifacts']['templates'] = {}
         for file_name in self.templates.keys():
@@ -1142,8 +1208,40 @@ class SubExperiment(AbstractExperiment):
         self.prevent_execution = True
 
     def __enter__(self):
+        self.check_experiment_exchange()
+
+        # 13.02.2023
+        # This fixes a rather serious bug, where the "analysis.py" file in the archive folder
+        # would not work at all for SubExperiments. This was because when imported rather than
+        # executed, the skipping mechanism of AbstractExperiment would immediately trigger for
+        # the SubExperiment such that it would not actually import the parent experiment and thus
+        # all the parameters defined only there would not be available to the analysis script
+        # leading to an error.
+
+        # Now what we do here is we make sure that we do import the parent module before we call
+        # the skipping functionality of AbstractExperiment, but only if we recognize that this
+        # experiment itself is being imported.
+        # Since we modify __name__ for the parent experiment as well that one will not actually
+        # execute either.
+        if self.glob["__name__"] != "__main__":
+            self.prevent_execution = True
+            self.analysis.skip = True
+
+            # This method will place the local dict "data" which contains the updated experiment information
+            # into the global experiment data exchange object keyed with the name of the experiment module's
+            # path, so that the parent experiment can access that data during it's __enter__ operation.
+            data = {'glob': {'__name__': '__import__'}}
+            self.experiment_exchange.request(self.experiment_path, data=data)
+
+            # This method will dynamically import the experiment given by self.experiment_path
+            module = self.dynamic_import_experiment()
+
+            experiment = getattr(module, '__experiment__')
+            self.update(experiment)
+
+        # Only after we did that we call the functionality of AbstractExperiment which would then
+        # ultimately trigger the context skip if it was truly being just imported.
         return super(SubExperiment, self).__enter__()
-        pass
 
     def get_clean_glob(self) -> t.Dict[str, t.Any]:
         """
@@ -1195,9 +1293,8 @@ class SubExperiment(AbstractExperiment):
         # path, so that the parent experiment can access that data during it's __enter__ operation.
         self.experiment_exchange.request(self.experiment_path, data=data)
 
-        spec = importlib.util.spec_from_file_location('experiment', self.experiment_path)
-        module = importlib.util.module_from_spec(spec)
-        spec.loader.exec_module(module)
+        # This method will dynamically import the experiment given by self.experiment_path
+        module = self.dynamic_import_experiment()
 
         # At experiment __enter__ EVERY experiment will save a reference to itself into it own module's
         # global variable called __experiment__. We access the experiment like that and use it to update
@@ -1207,6 +1304,13 @@ class SubExperiment(AbstractExperiment):
         experiment = getattr(module, '__experiment__')
         self.update(experiment)
 
+        # 13.02.2023
+        # This is needed ot make the "analysis.py" files expandable. We essentially prepend the analysis
+        # code which was already recorded by the execution of the parent experiment to the local analysis
+        # context manager so that any analysis code defined on this level will simply be appended to the
+        # end of already existing code
+        self.analysis.code_lines = experiment.analysis.code_lines
+
         # 20.01.2023
         # This fixes the issue where the snapshot of a sub experiment is not in fact executable, because
         # the base experiment upon which it extends does not exist in the same folder. So here we copy that
@@ -1214,6 +1318,24 @@ class SubExperiment(AbstractExperiment):
         destination_path = os.path.join(self.path, self.experiment_name)
         shutil.copy(self.experiment_path, destination_path)
 
+    def dynamic_import_experiment(self) -> 'ModuleType':
+        """
+        This method dynamically imports the base experiment file which is identified by the absolute path
+        given by ``self.experiment_path``. This also means that the top-level code of that module is
+        also EXECUTED within this method!
+        Returns the imported module object instance.
+        """
+        # 24.03.2023 - Added the derivation of the module name, and now we also add that uniquely named
+        # module to the sys.modules directory. This is NECESSARY to fix a bug where the "inspect"
+        # functionality would not work properly in SubExperiments!
+        module_name = os.path.basename(self.experiment_path).strip('.py')
+        spec = importlib.util.spec_from_file_location(module_name, self.experiment_path)
+        module = importlib.util.module_from_spec(spec)
+        sys.modules[module_name] = module
+        spec.loader.exec_module(module)
+
+        return module
+
 
 def run_experiment(experiment_path: str,
                    parameters: dict = {},
